#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include <string.h>
#include <stdio.h>
#include "SaBLExAPI_IncomingMsg_Common.h"
#include "SaBLExAPI_OutgoingMsg_Common.h"
#include "circularBuffer.h"
#include "uartProcessor.h"
#include "mspUart.h"
#include "SaBLExHost.h"
#include "SaBLExAPI_IncomingMsg_Peripheral.h"
#include "SaBLExAPI_OutgoingMsg_Peripheral.h"
#include "SaBLExAPI_Shared.h"
#include "SaBLExHost.h"

/*-----------------------------------------------------------*/

#define applicationSTACK_SIZE	        configMINIMAL_STACK_SIZE * 3
#define application_MAX_NUM_EVENTS      20
#define application_MAX_NUM_OTA_MSGS    10
#define SEND_LIGHT_DATA                 0
#define SEND_TEMP_DATA                  1
#define SEND_TILT_DATA                  2

/*-----------------------------------------------------------*/

typedef struct appEvt
{
    uint8_t u8Event;
    uint16_t u16Param;
    void* pvData;
}appEvt_t;

typedef struct otaMsg
{
    char* pcData;
    bool bIsAscii;
}otaMsg_t;

/*-----------------------------------------------------------*/

static uint8_t u8DataToSend = SEND_LIGHT_DATA;
static bool bAsleep = false;
static bool bWaitingToSleepAfterUartTx = false;
static bool bProcedureInProgress = false;
static QueueHandle_t xEventsForApp;
static QueueHandle_t xMsgsForOtaTx;
static uint8_t u8JsonId;
static bool bConnected = false;
static const char* pcAdvertisingMsg = "Advertising";
static const char* pcConnected = "Connected";
static const char* pcDisconnected = "Disconnected";
static const char* pcTransmissionPass = "OTA Tx Pass";
static const char* pcTransmissionFail = "OTA Tx Fail";
static void (*functionAfterModuleAwake)(void) = NULL;
static portTASK_FUNCTION_PROTO( vApplicationTask, pvParameters );

/*-----------------------------------------------------------*/

static void SaBLExHost_ProcessAppEvt(appEvt_t* pMsg);
static void SaBLExHost_BringModuleOutOfRst(void);
static void SaBLExHost_FirmwareCheck(results_QueryFirmwareVersionIn_t* results);
static void SaBLExHost_PrintMacAddress(char* pcMac);
static void SaBLExHost_SetLedBehavior(void);
static void SaBLExHost_SetAdvertisingParams(void);
static void SaBLExHost_EnableAdvertising(void);
static void SaBLExHost_SetConnectionParams(void);
static void SaBLExHost_EnableDevSensors(void);
static void SaBLExHost_SetLpmParams(void);
static void SaBLExHost_SetLpmMode(void);
static void SaBLExHost_SystemInitialized(void);
static void SaBLExHost_Connected(void);
static void SaBLExHost_Disconnected(void);
static void SaBLExHost_OtaDataReceived(uint8_t u8Id, params_ReceiveOtaDataIn_t* otaData);
static void SaBLExHost_QuerySensors(void);
static void SaBLExHost_SendSensorDataOta(results_QueryDevBoardSensorsIn_t* sensorData);
static void SaBLExHost_TransferOtaMessage(void);

/*-----------------------------------------------------------*/

void SaBLExHost_StartTask( UBaseType_t uxPriority)
{
    /* Create the queues used by the com test task. */
    xEventsForApp = xQueueCreate( application_MAX_NUM_EVENTS, ( unsigned portBASE_TYPE ) sizeof( appEvt_t ) );
    xMsgsForOtaTx = xQueueCreate( application_MAX_NUM_OTA_MSGS, ( unsigned portBASE_TYPE ) sizeof( otaMsg_t ) );

    xTaskCreate( vApplicationTask, "App", applicationSTACK_SIZE, NULL, uxPriority, (TaskHandle_t*) NULL );
}

bool SaBLExHost_Asleep( void )
{
    return bAsleep;
}

void SaBLExHost_SetAwake( void )
{
    bAsleep = false;
}


/* Used to enqueue events from other tasks */
void SaBLExHost_EnqueueEvt(uint8_t u8Event, uint16_t u16Param, void* pvData)
{
    appEvt_t appEvt;
    appEvt.u8Event = u8Event;
    appEvt.u16Param = u16Param;
    appEvt.pvData = pvData;
    xQueueSend( xEventsForApp, ( void * ) &appEvt, ( TickType_t ) 0 );
}

/* Used to enqueue events from ISRs */
void SaBLExHost_EnqueueEvtFromISR(uint8_t u8Event, uint16_t u16Param, void* pvData)
{
    appEvt_t appEvt;
    appEvt.u8Event = u8Event;
    appEvt.u16Param = u16Param;
    appEvt.pvData = pvData;
    xQueueSendFromISR( xEventsForApp, ( void * ) &appEvt, ( TickType_t ) 0 );
}

static portTASK_FUNCTION( vApplicationTask, pvParameters )
{
	/* Just to stop compiler warnings. */
	( void ) pvParameters;

    SaBLExHost_BringModuleOutOfRst();
    
	for( ;; )
	{
        appEvt_t appEvt;
        if( xQueueReceive( xEventsForApp, &( appEvt ), portMAX_DELAY ))
        {
            SaBLExHost_ProcessAppEvt(&appEvt);
        }
    }
} /*lint !e715 !e818 pvParameters is required for a task function even if it is not referenced. */


static void SaBLExHost_ProcessAppEvt(appEvt_t* pMsg)
{
    xLcdQueueMessage msg;
    
    switch(pMsg->u8Event)
    {
        case SABLE_X_RESET_NOTIFICATION_EVT:
        SetExpectedMsgIdAndFunction(u8JsonId, QueryFirmwareVersionIn);
        QueryFirmwareVersionOut (u8JsonId++);
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_QUERY_FIRMWARE_RESP_EVT:
        SaBLExHost_FirmwareCheck(pMsg->pvData);
        break;
/*-----------------------------------------------------------*/        
        case SABLE_X_QUERY_MAC_EVT:
        SetExpectedMsgIdAndFunction(u8JsonId, QueryMacAddressIn);
        QueryMacAddressOut (u8JsonId++);        
        break;
/*-----------------------------------------------------------*/        
        case SABLE_X_QUERY_MAC_RESP_EVT:
        SaBLExHost_PrintMacAddress(pMsg->pvData);
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_RESTORE_FACTORY_DEFAULTS_EVT:
        SetExpectedMsgIdAndFunction(u8JsonId, ResetToFactoryDefaultsIn);
        ResetToFactoryDefaultsOut (u8JsonId++);        
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_RESTORE_FACTORY_DEFAULTS_RESP_EVT:
        SaBLExHost_SetLedBehavior();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SET_PERIPH_LED_PARAMS_RESP_EVT:
        SaBLExHost_SetAdvertisingParams();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SET_ADVERTISING_PARAMS_RESP_EVT:
        SaBLExHost_EnableAdvertising();
        break;
        case SABLE_X_SET_ADVERTISING_ENABLE_RESP_EVT:
        SaBLExHost_SetConnectionParams();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SET_CONNECTION_PARAMS_RESP_EVT:
        SaBLExHost_EnableDevSensors();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SET_DEV_BOARD_ENABLE_RESP_EVT:
        SaBLExHost_SetLpmParams();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SET_LPM_PARAMS_RESP_EVT:
        SaBLExHost_SetLpmMode();
        break;
/*-----------------------------------------------------------*/        
        case SABLE_X_SET_LPM_MODE_RESP_EVT:
        SaBLExHost_SystemInitialized();
        bAsleep = true;        
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_LINK_STATUS_EVT:
        bConnected = pMsg->u16Param;
        
        if(bConnected)
        {
            SaBLExHost_Connected();
        }
        else
        {
            SaBLExHost_Disconnected();
        }
        bAsleep = true;
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_WAKE_HOST_EVT:
        HostAwakeOut();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_RECEIVE_OTA_DATA_EVT:
        SaBLExHost_OtaDataReceived(pMsg->u16Param, pMsg->pvData);
        bWaitingToSleepAfterUartTx = true;
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_UART_TX_COMPLETE:
        if(bWaitingToSleepAfterUartTx)
        {
           bWaitingToSleepAfterUartTx = false;
           /* Delay a brief moment prior to sleeping so last byte can leave
              the shift register */
           const TickType_t xDelay = 10;
           vTaskDelay( xDelay );
           bAsleep = true; 
        }
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SEND_SESNSOR_DATA_OTA:
        if(bConnected  && !bProcedureInProgress)
        {
            bProcedureInProgress = true;
            functionAfterModuleAwake = SaBLExHost_QuerySensors;
            WakeupModuleOut();
        }
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_MODULE_AWAKE_EVT:
        if(functionAfterModuleAwake != NULL)
        {
            functionAfterModuleAwake();
            functionAfterModuleAwake = NULL;
        }
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_QUERY_DEV_SENSORS_RESP_EVT:
        SaBLExHost_SendSensorDataOta(pMsg->pvData);
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_TRANSFER_OTA_EVT:
        SaBLExHost_TransferOtaMessage();
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_SEND_OTA_DATA_RESP_EVT:
        bAsleep = true;
        break;
/*-----------------------------------------------------------*/
        case SABLE_X_RF_ACK_EVT:
        bProcedureInProgress = false;
        if(pMsg->u16Param)
        {
           msg.pcMessage = pvPortMalloc(strlen(pcTransmissionFail)+1);     
           memcpy(msg.pcMessage, pcTransmissionFail, strlen(pcTransmissionFail)+1);
        }
        else
        {
           msg.pcMessage = pvPortMalloc(strlen(pcTransmissionPass)+1);     
           memcpy(msg.pcMessage, pcTransmissionPass, strlen(pcTransmissionPass)+1);
        }
        xQueueSend( xLCDQueue, &msg, portMAX_DELAY );
        bAsleep = true;
        break;
/*-----------------------------------------------------------*/
        default:
        break;

    }
    /* Free queued msg's payload if it exists. */
    if(pMsg->pvData != NULL)
    {
        vPortFree(pMsg->pvData);
    }
}

static void SaBLExHost_BringModuleOutOfRst(void)
{
    P1OUT |= 0x04;
}

static void SaBLExHost_FirmwareCheck(results_QueryFirmwareVersionIn_t* results)
{
    if((strcmp(results->Partnumber, "480-0038") == 0) &&
       (strcmp(results->Desc, "Serial-to-BLE, Peripheral") == 0))
    {
        SaBLExHost_EnqueueEvt(SABLE_X_QUERY_MAC_EVT, 0, NULL);
    }
    else
    {
        /* Unsupported firmware */
        for(;;);
    }
       
    vPortFree(results->Desc);
    vPortFree(results->Partnumber);
}

static void SaBLExHost_PrintMacAddress(char* pcMac)
{
    xLcdQueueMessage msg;
    msg.pcMessage = pvPortMalloc(strlen(pcMac)+1);
    memcpy(msg.pcMessage, pcMac, strlen(pcMac)+1);
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY );   
    SaBLExHost_EnqueueEvt(SABLE_X_RESTORE_FACTORY_DEFAULTS_EVT, 0, NULL);
}

static void SaBLExHost_SetLedBehavior(void)
{
    params_SetPeripheralLedBehaviorOut_t temp;
    temp.AdvertisingPeriod = 1000;   /* Blink Advertising LED every 500ms when advertising. */
    temp.HeartbeatPeriod = 3000;     /* Blink Heartbeat LED every 500ms. */
    temp.UartActivity = true;       /* Blink UART LED when Tx'ing or Rx'ing valid Json msg. */
    temp.SaveToNv = false;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */
    SetExpectedMsgIdAndFunction(u8JsonId, SetPeripheralLedBehaviorIn);
    /* Send message to set the led behavior, increment the id after sending. */    
    SetPeripheralLedBehaviorOut ( u8JsonId++, &temp);    
}

static void SaBLExHost_SetAdvertisingParams(void)
{
    params_SetAdvertisingParamsOut_t temp;
    temp.AdvertisingChannels = 7;
    
    /* passed as ascii, but format is hex
       this sets the local name short to SaBLE-x */
    char advData[] =
    {
      '5','3','6','1','4','2','4','C','4','5','2','D','7','8',0 //null terminate
    };
    temp.AdvertisingData = advData;
    
    
    temp.AdvertisingInterval = 200;
    temp.DiscoverableMode = 1;
    temp.SaveToNv = false;

    /* passed as ascii, but format is hex
       this sets the local name complete to MSP430-Demo */
    char scanData[] = 
    {
        '4','D','5','3','5','0','3','4','3','3','3','0','2','D','4','4','6','5','6','D','6','F',0 //null terminate
    };
    
    temp.ScanResponseData = scanData;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */    
    SetExpectedMsgIdAndFunction(u8JsonId, SetAdvertisingParamsIn);
    /* Send message to set the advertising params, increment the id after sending. */     
    SetAdvertisingParamsOut (u8JsonId++, &temp);
}

static void SaBLExHost_EnableAdvertising(void)
{
    params_SetAdvertisingEnableOut_t temp;
    temp.AdvertisingEnable = true;
    temp.SaveToNv = false;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */    
    SetExpectedMsgIdAndFunction(u8JsonId, SetAdvertisingEnableIn);
    /* Send message to set the advertising enable, increment the id after sending. */       
    SetAdvertisingEnableOut (u8JsonId++, &temp);
}

static void SaBLExHost_SetConnectionParams(void)
{
    params_SetConnectionParamsOut_t temp;
    temp.MaxConnectionInterval = 6;
    temp.MinConnectionInterval = 6;
    temp.SlaveLatency = 0;
    temp.SupervisionTimeout = 1000;
    temp.SaveToNv = false;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */    
    SetExpectedMsgIdAndFunction(u8JsonId, SetConnectionParamsIn);
    /* Send message to set the connection params, increment the id after sending. */      
    SetConnectionParamsOut (u8JsonId++, &temp);
}

static void SaBLExHost_EnableDevSensors(void)
{
    params_SetDevBoardSensorsEnableOut_t temp;
    temp.SaveToNv = false;
    temp.Enable = true;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */
    SetExpectedMsgIdAndFunction(u8JsonId, SetDevBoardSensorsEnableIn);
    /* Send message to set the dev board sensors enable, increment the id after sending. */     
    SetDevBoardSensorsEnableOut ( u8JsonId++, &temp); 
}

static void SaBLExHost_SetLpmParams(void)
{
    params_SetLpmParamsOut_t temp;
    temp.SaveToNv = false;
    temp.SleepTime = 300;
    temp.WakeHost = true;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */    
    SetExpectedMsgIdAndFunction(u8JsonId, SetLpmParamsIn);
    /* Send message to set the lpm params, increment the id after sending. */    
    SetLpmParamsOut ( u8JsonId++, &temp);     
}

static void SaBLExHost_SetLpmMode(void)
{
    /* Explicitly put the module to sleep.  This sets a flag in the module code so
       the module instantly sleeps after any RF rx or tx message.  Otherwise the module
       will remain awake for the duration of the sleep time set above. */
    params_ModuleGoToLpmOut_t temp;
    temp.Mode = 0;
    /* Set the expected id and the function to execute upon receiving the result message with that id. */    
    SetExpectedMsgIdAndFunction(u8JsonId, ModuleGoToLpmIn);
    /* Send message to set the lpm mode, increment the id after sending. */    
    ModuleGoToLpmOut ( u8JsonId++, &temp);      
}

static void SaBLExHost_SystemInitialized(void)
{
    xLcdQueueMessage msg;
    msg.pcMessage = pvPortMalloc(strlen(pcAdvertisingMsg)+1);
    memcpy(msg.pcMessage, pcAdvertisingMsg, strlen(pcAdvertisingMsg)+1);
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY );    
}

static void SaBLExHost_Connected(void)
{
    xLcdQueueMessage msg;
    msg.pcMessage = pvPortMalloc(strlen(pcConnected)+1);
    memcpy(msg.pcMessage, pcConnected, strlen(pcConnected)+1);
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY );  
}


static void SaBLExHost_Disconnected(void)
{
    xLcdQueueMessage msg;
    msg.pcMessage = pvPortMalloc(strlen(pcDisconnected)+1);
    memcpy(msg.pcMessage, pcDisconnected, strlen(pcDisconnected)+1);
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY ); 

    msg.pcMessage = pvPortMalloc(strlen(pcAdvertisingMsg)+1);
    memcpy(msg.pcMessage, pcAdvertisingMsg, strlen(pcAdvertisingMsg)+1);
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY );    
}

static void SaBLExHost_OtaDataReceived(uint8_t u8Id, params_ReceiveOtaDataIn_t* otaData)
{
    xLcdQueueMessage msg;
    
    if(otaData->DataAscii != NULL)
    {
        msg.pcMessage = pvPortMalloc(strlen(otaData->DataAscii)+1);
        memcpy(msg.pcMessage, otaData->DataAscii, strlen(otaData->DataAscii)+1);
        vPortFree(otaData->DataAscii);
    }
    else
    {
        msg.pcMessage = pvPortMalloc(strlen(otaData->DataHex)+1);
        memcpy(msg.pcMessage, otaData->DataHex, strlen(otaData->DataHex)+1);
        vPortFree(otaData->DataHex);
    }    
    xQueueSend( xLCDQueue, &msg, portMAX_DELAY );  
    
    /* Send result (UART acknowledgement) to the module so it doesn't resend msg. */    
    results_ReceiveOtaDataOut_t temp;
    temp.Status = 0;
    ReceiveOtaDataOut (u8Id, &temp);
}

static void SaBLExHost_QuerySensors(void)
{
    /* Set the expected id and the function to execute upon receiving the result message with that id. */
    SetExpectedMsgIdAndFunction(u8JsonId, QueryDevBoardSensorsIn);
    /* Send message to query the dev board sensor, increment the id after sending. */      
    QueryDevBoardSensorsOut ( u8JsonId++);    
}

static void SaBLExHost_SendSensorDataOta(results_QueryDevBoardSensorsIn_t* sensorData)
{
    
    otaMsg_t otaMsg;
    otaMsg.bIsAscii = true;    
    
    if(u8DataToSend == SEND_LIGHT_DATA)
    {
        double dLux = ((double)sensorData->LightSensor) / 1000.0;
        char* pcLux = pvPortMalloc(21);
        memset(pcLux, 0, 21);
        pcLux[0] = 'L';
        pcLux[1] = 'U';
        pcLux[2] = 'X';
        pcLux[3] = ':';
        pcLux[4] = ' ';
        snprintf(pcLux+5,15,"%f",dLux);
        otaMsg.pcData = pcLux;
        u8DataToSend = SEND_TEMP_DATA;
    }
    else if(u8DataToSend == SEND_TEMP_DATA)
    {
        double dTemp = ((double)sensorData->TempSensor) / 1000.0;
        char* pcTemp = pvPortMalloc(21);
        memset(pcTemp, 0, 21);
        pcTemp[0] = 'T';
        pcTemp[1] = 'E';
        pcTemp[2] = 'M';
        pcTemp[3] = 'P';
        pcTemp[4] = ':';
        pcTemp[5] = ' ';
        snprintf(pcTemp+6,14,"%f",dTemp);
        otaMsg.pcData = pcTemp;
        u8DataToSend = SEND_TILT_DATA;
    }
    else
    {
        char* pcTilt = pvPortMalloc(10);
        memset(pcTilt, 0, 10);
        pcTilt[0] = 'T';
        pcTilt[1] = 'I';
        pcTilt[2] = 'L';
        pcTilt[3] = 'T';
        pcTilt[4] = ':';
        pcTilt[5] = ' ';
        snprintf(pcTilt+6,3,"%d",sensorData->AccelerometerTilt);
        otaMsg.pcData = pcTilt;
        u8DataToSend = SEND_LIGHT_DATA;
    }
    /* Add to queue and spark event to initiate over the air transmission */
    xQueueSend( xMsgsForOtaTx, ( void * ) &otaMsg, ( TickType_t ) 0 );        
    SaBLExHost_EnqueueEvt(SABLE_X_TRANSFER_OTA_EVT, 0, NULL);
    
}

static void SaBLExHost_TransferOtaMessage(void)
{
    otaMsg_t otaMsg;

    if( xQueueReceive( xMsgsForOtaTx, &( otaMsg ), ( TickType_t ) 0 ))
    {
        params_SendOtaDataOut_t temp;
        
        if(otaMsg.bIsAscii)
        {
            temp.DataAscii = otaMsg.pcData;
            temp.DataHex = NULL;
        }
        else
        {
            temp.DataAscii = NULL;
            temp.DataHex = otaMsg.pcData;            
        }
        /* Set the expected id and the function to execute upon receiving the result message with that id. */
        SetExpectedMsgIdAndFunction(u8JsonId, SendOtaDataIn);
        /* Send message to set send over the air data, increment the id after sending. */ 
        SendOtaDataOut (u8JsonId++, &temp);
    }    
}
