#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"
#include "circularBuffer.h"
#include "mspUart.h"
#include <stdlib.h>
#include <stdbool.h>
#include "SaBLExHost.h"

/*-----------------------------------------------------------*/

/* Misc. constants. */
#define serNO_BLOCK				( ( TickType_t ) 0 )

/*-----------------------------------------------------------*/

static circular_buffer_t* rxBuf;

/*-----------------------------------------------------------*/

static TaskHandle_t xTaskToNotify = NULL;
/* The queue used to hold characters waiting transmission. */
static QueueHandle_t xCharsForTx;
/* Save the value of the baud rate so it can be used when initializing the UART after lpm wakeup. */
static unsigned long ulBaudRateCountForReuse;

/*-----------------------------------------------------------*/

void MspUart_Init( unsigned long ulWantedBaud, unsigned portBASE_TYPE uxTxQueueLength )
{
	/* Initialise the hardware. */

	/* Generate the baud rate constants for the wanted baud rate. */
    ulBaudRateCountForReuse = configCPU_CLOCK_HZ / ulWantedBaud;
	unsigned long ulBaudRateCount = ulBaudRateCountForReuse;

	portENTER_CRITICAL();
	{
		/* Create the queues used by the com test task. */
		xCharsForTx = xQueueCreate( uxTxQueueLength, ( unsigned portBASE_TYPE ) sizeof( char ) );

		/* Reset UART. */
		UCA2CTL1 |= UCSWRST;

        P9SEL |= 0x30;
        
		/* Use SMCLK. */
		UCA2CTL1 = UCSSEL0 | UCSSEL1;
		
		/* Setup baud rate low byte. */
		UCA2BR0 = ( unsigned char ) ( ulBaudRateCount & ( unsigned long ) 0xff );

		/* Setup baud rate high byte. */
		ulBaudRateCount >>= 8UL;
		UCA2BR1 = ( unsigned char ) ( ulBaudRateCount & ( unsigned long ) 0xff );

		/* Enable interrupts. */
		UCA2IE |= UCRXIE;
		
		/* Take out of reset. */
		UCA2CTL1 &= ~UCSWRST;
	}
	portEXIT_CRITICAL();
}

/*-----------------------------------------------------------*/

void MspUart_Disable( void )
{
	portENTER_CRITICAL();
	{
       
        /* Disable UART pin and put to input pulled up */
        P9SEL &= ~0x20;    

        P9DIR &= ~0x20;
        P9REN |= 0x20; 
        
		/* Reset UART. */
		UCA2CTL1 |= UCSWRST;
    }
	portEXIT_CRITICAL();    
}

/*-----------------------------------------------------------*/

void MspUart_Enable( void )
{
unsigned long ulBaudRateCount = ulBaudRateCountForReuse;
	portENTER_CRITICAL();
	{
        
		/* Reset UART. */
		UCA2CTL1 |= UCSWRST;

        P9SEL |= 0x20;
        
		/* Use SMCLK. */
		UCA2CTL1 = UCSSEL0 | UCSSEL1;
		
		/* Setup baud rate low byte. */
		UCA2BR0 = ( unsigned char ) ( ulBaudRateCount & ( unsigned long ) 0xff );

		/* Setup baud rate high byte. */
		ulBaudRateCount >>= 8UL;
		UCA2BR1 = ( unsigned char ) ( ulBaudRateCount & ( unsigned long ) 0xff );

		/* Enable interrupts. */
		UCA2IE |= UCRXIE;
		
		/* Take out of reset. */
		UCA2CTL1 &= ~UCSWRST;
    }
	portEXIT_CRITICAL();    
}

/*-----------------------------------------------------------*/

void MspUart_RegisterTaskHandleAndBuffers(circular_buffer_t* rxBuffer )
{
    rxBuf = rxBuffer;
    xTaskToNotify = xTaskGetCurrentTaskHandle();
}

/*-----------------------------------------------------------*/

signed portBASE_TYPE MspUart_Write( char* pcBuffer, uint16_t u16Len )
{
signed portBASE_TYPE xReturn;

	/* Send the next characters to the queue of characters waiting transmission,
	then enable the UART Tx interrupt, just in case UART transmission has already
	completed and switched itself off. */
    
    for(uint16_t i = 0; i<u16Len; i++)
    {
        xReturn = xQueueSend( xCharsForTx, &pcBuffer[i], serNO_BLOCK );
        
        if(xReturn != 1)
        {
            return xReturn;
        }
    }
    
    /* Enable Tx interrupt */
	UCA2IE |= UCTXIE;

	return xReturn;
}
/*-----------------------------------------------------------*/

#pragma vector=USCI_A2_VECTOR
static __interrupt void prvUSCI_A2_ISR( void )
{
char cChar;
portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	if( ( UCA2IFG & UCRXIFG ) != 0 )
	{
        cChar = (char)UCA2RXBUF;

        /* Get the character from the UART and post it on the queue of Rxed
        characters. */

        cb_push_back(rxBuf, &cChar);
        
        eTaskState eState = eTaskGetStateFromISR(xTaskToNotify);
        if(eState == eSuspended || eState == eReady)
        {
            /* Notify app that transmission of Tx msg complete (used for sleep cases). */
            vTaskNotifyGiveFromISR( xTaskToNotify, &xHigherPriorityTaskWoken );
        }        
	}
	
	/* If there is a Tx interrupt pending and the tx interrupts are enabled. */
	if( ( UCA2IFG & UCTXIFG ) != 0 )
	{
		/* The previous character has been transmitted.  See if there are any
		further characters waiting transmission. */
		if( xQueueReceiveFromISR( xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
		{
			/* There was another character queued - transmit it now. */
			UCA2TXBUF = cChar;
		}
		else
		{
			/* There were no other characters to transmit - disable the Tx
			interrupt. */
			UCA2IE &= ~UCTXIE;
            // This event is used so the app knows a TX message was sucessfully transmitted
            // Its main purpose is to indicate that it is safe to go to LPM
            SaBLExHost_EnqueueEvtFromISR(SABLE_X_UART_TX_COMPLETE, 0, NULL);
		}
	}
	
	/* If writing to a queue caused a task to unblock, and the unblocked task
	has a priority equal to or above the task that this interrupt interrupted,
	then lHigherPriorityTaskWoken will have been set to pdTRUE internally within
	xQueuesendFromISR(), and portEND_SWITCHING_ISR() will ensure that this
	interrupt returns directly to the higher priority unblocked task.
	
	THIS MUST BE THE LAST THING DONE IN THE ISR. */	
	portYIELD_FROM_ISR( xHigherPriorityTaskWoken );
}